
/* tcptb
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2004 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pcap.h>
#include <tcptb.h>
#include <tcptb-int.h>

pcap_t *tcptb_dump;
unsigned short tcptb_first = TRUE;
char *tcptb_filter = NULL;


/* Spit a version number */

char *tcptb_version ()
{
    return (TCPTB_VERSION);
}

/* Returns somewhat descriptive text about the most recent tcptb error. */

char *tcptb_error ()
{
    return (tcptb_err_msg);
}

/* Open the given packet trace.  Caveats:

   We only care about TCP packets.
   If a trace is already open we silently close it (but, users should not
   be dealing with the file directly anyway).
 */

int tcptb_open (char *filename)
{
    char errbuf [PCAP_ERRBUF_SIZE];
    char buffer [PCAP_ERRBUF_SIZE + TCPTB_ERR_SIZE];
    struct bpf_program bpfp;
    char filter_string [INT_BUFFER_SIZE];

    if (tcptb_fatal_error)
	return (-1);
    if (tcptb_dump != NULL)
	tcptb_close (tcptb_dump);
    tcptb_first = TRUE;
    if ((tcptb_dump = pcap_open_offline (filename,errbuf)) == NULL)
    {
	sprintf (buffer,"tcptb_open (): %s", errbuf);
	tcptb_set_error (buffer,TRUE);
	return (-1);
    }
    if (tcptb_filter != NULL)
	sprintf (filter_string,"tcp and (%s)", tcptb_filter);
    else
	sprintf (filter_string,"tcp");
    if (pcap_compile (tcptb_dump,&bpfp,filter_string,1,0) < 0)
    {
	tcptb_set_error ("can't compile packet filter",TRUE);
	return (-1);
    }
    if (pcap_setfilter (tcptb_dump,&bpfp) < 0)
    {
	tcptb_set_error ("can't set packet filter",TRUE);
	return (-1);
    }
    if (pcap_datalink (tcptb_dump) != DLT_EN10MB)
    {
	tcptb_set_error ("bad data link type",TRUE);
	return (-1);
    }
    return (0);
}

/* Close the packet trace. */

void tcptb_close ()
{
    if (tcptb_fatal_error)
	return;
    if (tcptb_dump == NULL)
	return;
    pcap_close (tcptb_dump);
    tcptb_dump = NULL;
}

/* Read in all the packets from the trace. */

int tcptb_slurp_trace ()
{
    if (tcptb_fatal_error || (tcptb_dump == NULL))
    {
	tcptb_set_error ("trace file not open",FALSE);
	return (-1);
    }
    if (tcptb_read_pkts (tcptb_dump) < 0)
	return (-1);
    tcptb_first = FALSE;
    return (0);
}

/* Grab the next connection from the trace.  If we have not yet processed 
   the trace, it will be slurped.

   NOTE: The routine that actually find the "next" connection always 
   returns the same connection unless you run the next routine (to free
   a connection).  This is pretty dumb, but unfortunately getting the 
   "next" connection is not as easy as it might seem (given connections
   coming and going).  So, I punted. */

struct tcptb_cnn *tcptb_next_cnn ()
{
    struct tcptb_cnn *cnn;

    if (tcptb_fatal_error)
	return (NULL);
    if (tcptb_first)
    {
	if (tcptb_slurp_trace () < 0)
	    return (NULL);
    }
    cnn = tcptb_get_cnn ();
    return (cnn);
}

/* "Free" a connection. */

int tcptb_free_cnn (struct tcptb_cnn *cnn)
{
    if (tcptb_fatal_error)
	return (-1);
    return (tcptb_nuke_cnn (cnn));
}

/* Sending a zero flag (the default) causes the tcptb to lop off any
   TCP payload that ended up in the trace file when storing the packets
   in memory.  This saves memory when you only care about the headers.
   But, if you care about the application data this sort of sucks. */

void tcptb_save_whole_packet (unsigned short flag)
{
    tcptb_keep_whole_packet = flag;
}

/* Sending a zero flag (the default) causes the tcptb to throw away
   the ethernet header of each packet processed.  This saves memory 
   in the common case of not requiring the ethernet headers when 
   analyzing TCP.  If you care about the ethernet headers send a 
   non-zero value and they will be kept. */

void tcptb_save_ether_hdr (unsigned short flag)
{
    tcptb_keep_ether_hdr = flag;
}

/* This allows you to control the size of the hash table used to track
   connections.  The default has worked well for us for web traces on the
   order of 1GB -- hence, could be wasteful for short traces of only a
   few connections.  (Remember, use a prime for the hash table size!) */

int tcptb_set_hash_table_size (unsigned long sz)
{
    if (tcptb_table_init)
    {
	tcptb_set_error ("error: hash table already created; can't resize",
			 FALSE);
	return (-1);
    }
    tcptb_hash_table_size = sz;
    return (0);
}


int tcptb_set_bpf_filter (char *filter)
{
    if (tcptb_filter != NULL)
    {
	tcptb_set_error ("can't set filter multiple times",FALSE);
	return (-1);
    }
    if ((filter == NULL) || !strlen (filter))
    {
	tcptb_set_error ("no filter passed",FALSE);
	return (-1);
    }
    if ((tcptb_filter = (char *)malloc (strlen (filter) + 1)) == NULL)
    {
	tcptb_set_error ("cannot allocate memory for filter",FALSE);
	return (-1);
    }
    strcpy (tcptb_filter,filter);
    return (0);
}
