
/* tcptb
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2004 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcptb.h>
#include <tcptb-int.h>

struct tcptb_cnn **tcptb_cnn_table;
unsigned long tcptb_hash_table_size = DEFAULT_HASH_TBL_SIZE;
unsigned short tcptb_table_init = FALSE;


int tcptb_cnn_table_init ()
{
    unsigned long tbl_size;
    void *mem;

    tbl_size = tcptb_hash_table_size * sizeof (struct tcptb_cnn *);
    mem = (void *)tcptb_malloc (tbl_size);
    tcptb_cnn_table = (struct tcptb_cnn **)mem;
    if (tcptb_cnn_table == NULL)
	return (-1);
    bzero (tcptb_cnn_table,tbl_size);
    tcptb_table_init = TRUE;
    return (0);
}


unsigned int tcptb_hash_adjust (array, size, hash)
char *array;
unsigned int size, hash;
{
    int i;
    unsigned int h;

    h = hash;
    if (array != NULL)
	for (i = 0; i < size; i++)
	    h = (h * HASH_MULTIPLIER) + array [i];
    return (h);
}


unsigned int tcptb_cnn_hash (src_ip, dst_ip, src_port, dst_port)
unsigned long src_ip, dst_ip;
unsigned short src_port, dst_port;
{
    unsigned int hash = 0;

    if ((src_port < dst_port) || ((src_port == dst_port) &&
				  (src_ip < dst_ip)))
    {
        hash = tcptb_hash_adjust ((char *)&src_ip,IP_ADDR_SZ,hash);
        hash = tcptb_hash_adjust ((char *)&dst_ip,IP_ADDR_SZ,hash);
        hash = tcptb_hash_adjust ((char *)&src_port,TCP_PORT_SZ,hash);
        hash = tcptb_hash_adjust ((char *)&dst_port,TCP_PORT_SZ,hash);
    } 
    else 
    {
        hash = tcptb_hash_adjust ((char *)&dst_ip,IP_ADDR_SZ,hash);
        hash = tcptb_hash_adjust ((char *)&src_ip,IP_ADDR_SZ,hash);
        hash = tcptb_hash_adjust ((char *)&dst_port,TCP_PORT_SZ,hash);
        hash = tcptb_hash_adjust ((char *)&src_port,TCP_PORT_SZ,hash);
    }
    hash = hash % tcptb_hash_table_size;
    return (hash);
}


struct tcptb_cnn *tcptb_cnn_known (src_ip, dst_ip, src_port, dst_port)
unsigned long src_ip, dst_ip;
unsigned short src_port, dst_port;
{
    struct tcptb_cnn *p;
    unsigned int hash;

    if (!tcptb_table_init)
	return (NULL);
    hash = tcptb_cnn_hash (src_ip, dst_ip, src_port, dst_port);
    for (p = tcptb_cnn_table [hash]; p != NULL; p = p->next)
    {
	if ((src_ip == p->ip [0]) && (dst_ip == p->ip [1]) && 
	    (src_port == p->port [0]) && (dst_port == p->port [1]))
	    break;
	if ((dst_ip == p->ip [0]) && (src_ip == p->ip [1]) && 
	    (dst_port == p->port [0]) && (src_port == p->port [1]))
	    break;
    }
    return (p);
}


struct tcptb_cnn *tcptb_new_cnn (src_ip, dst_ip, src_port, dst_port)
unsigned long src_ip, dst_ip;
unsigned short src_port, dst_port;
{
    struct tcptb_cnn *newnode;
    unsigned int hash;

    if (!tcptb_table_init)
    {
	if (tcptb_cnn_table_init () < 0)
	{
	    tcptb_set_error ("fatal error: cannot allocate hash table",TRUE);
	    return (NULL);
	}
    }
    newnode = (struct tcptb_cnn *)tcptb_malloc (sizeof (struct tcptb_cnn));
    if (newnode == NULL)
    {
	tcptb_set_error ("fatal error: cannot allocate memory for connection",
			 TRUE);
	return (NULL);
    }
    bzero (newnode,sizeof (struct tcptb_cnn));
    hash = tcptb_cnn_hash (src_ip, dst_ip, src_port, dst_port);
    if (tcptb_cnn_table [hash] != NULL)
	newnode->next = tcptb_cnn_table [hash];
    tcptb_cnn_table [hash] = newnode;
    newnode->ip [0] = src_ip;
    newnode->ip [1] = dst_ip;
    newnode->port [0] = src_port;
    newnode->port [1] = dst_port;
    return (newnode);
}


/* This routine is really dumb.  Really dumb.  It will always return the
   same	connection unless the user of the library nukes the connections 
   as they are processed. */

struct tcptb_cnn *tcptb_get_cnn ()
{
    static unsigned int last_pos = 0;
    struct tcptb_cnn *cnn;

    if (!tcptb_table_init)
    {
	tcptb_set_error ("no connection table",FALSE);
	return (NULL);
    }
    if ((cnn = tcptb_get_restart_cnn ()) != NULL)
	return (cnn);
    for (; last_pos < tcptb_hash_table_size; last_pos++)
	if (tcptb_cnn_table [last_pos] != NULL)
	    return (tcptb_cnn_table [last_pos]);
    return (NULL);
}


int tcptb_nuke_cnn (cnn)
struct tcptb_cnn *cnn;
{
    struct tcptb_cnn *p, *q = NULL;
    unsigned int hash;

    if (!tcptb_table_init)
    {
	tcptb_set_error ("can't free from an unitialized connection table",
			 FALSE);
	return (-1);
    }
    if (cnn == NULL)
    {
	tcptb_set_error ("no connection to free given",FALSE);
	return (-2);
    }
    hash = tcptb_cnn_hash (cnn->ip [0], cnn->ip [1], cnn->port [0], 
			   cnn->port [1]);
    if ((p = tcptb_cnn_table [hash]) == NULL)
    {
	tcptb_set_error ("no entry in table for given connection",FALSE);
	return (-3);
    }
    while (p != NULL)
    {
	if (p == cnn)
	{
	    if (q == NULL)
		tcptb_cnn_table [hash] = cnn->next;
	    else
		q->next = p->next;
	    return (0);
	}
	q = p;
	p = p->next;
    }
    tcptb_set_error ("no entry in table for given connection",FALSE);
    return (-3);
}


#ifdef DEBUG
void tcptb_dump_hash_table_stats ()
{
    struct tcptb_cnn *p;
    unsigned int i;
    unsigned long c, t = 0;

    for (i = 0; i < tcptb_hash_table_size; i++)
    {
	for (p = tcptb_cnn_table [i], c = 0; p != NULL; p = p->next, c++, t++);
	fprintf (stdout,"bucket %d entries %ld\n", i, c);
    }
    fprintf (stdout,"total cnns %ld\n", t);
}
#endif /* DEBUG */
