
/* tcptb
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2004 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <tcptb.h>
#include <tcptb-int.h>

struct tcptb_pkt *tcptb_next_pkt_builtin ();

struct tcptb_pkt *(*tcptb_next_pkt_func)() = tcptb_next_pkt_builtin;
struct tcptb_cnn *tcptb_restart_cnn = NULL;
struct tcptb_pkt *tcptb_restart_pkt = NULL;
struct tcptb_pkt *tcptb_current_pkt = NULL;


void tcptb_set_restart_point (struct tcptb_cnn *cnn, struct tcptb_pkt *pkt)
{
    tcptb_restart_cnn = cnn;
    tcptb_restart_pkt = pkt;
}


struct tcptb_cnn *tcptb_get_restart_cnn ()
{
    return (tcptb_restart_cnn);
}


struct tcptb_pkt *tcptb_get_restart_pkt ()
{
    return (tcptb_restart_pkt);
}


void tcptb_set_next_pkt_function (struct tcptb_pkt *(*func)())
{
    tcptb_next_pkt_func = func;
}


struct tcptb_pkt *tcptb_next_pkt_builtin (struct tcptb_cnn *cnn,
					  struct tcptb_pkt *pkt)
{
    static unsigned short fin_or_rst_seen = FALSE;
    static unsigned short syn_seen = FALSE;
    static unsigned long syn_seqno;
    struct tcphdr *tcph;

    if ((cnn == NULL) || (pkt == NULL))
    {
	tcptb_set_error ("no connection or packet given to move from",FALSE);
	return (NULL);
    }
    if (((cnn == tcptb_get_restart_cnn ()) && 
	 (pkt == tcptb_get_restart_pkt ())) || (pkt == cnn->pkts_head))
    {
	/* first packet of a new connection */
	fin_or_rst_seen = FALSE;
	tcptb_set_restart_point (NULL,NULL);
	if (((tcph = tcptb_tcphdr (pkt)) != NULL) && (tcptb_is_syn (tcph)))
	{
	    syn_seqno = tcph->th_seq;
	    syn_seen = TRUE;
	}
	else
	    syn_seen = FALSE;
    }
    if ((pkt = pkt->next) == NULL)
    {
	fin_or_rst_seen = FALSE;
	tcptb_set_restart_point (NULL,NULL);
	return (NULL);
    }
    /* check whether this cnn is the same as the last cnn */
    if ((tcph = tcptb_tcphdr (pkt)) == NULL)
    {
	tcptb_set_error ("can't find TCP header",FALSE);
	return (NULL);
    }
    if ((tcptb_is_syn (tcph) && 
	 (tcptb_is_fin (tcph) || tcptb_is_rst (tcph))) ||
	(tcptb_is_fin (tcph) && tcptb_is_rst (tcph)))
	return (pkt);
    if (tcptb_is_fin (tcph) || tcptb_is_rst (tcph))
	fin_or_rst_seen = TRUE;
    if ((fin_or_rst_seen && tcptb_is_syn (tcph)) &&
	(!syn_seen || (syn_seen && (syn_seqno != tcph->th_seq))))
    {
	tcptb_set_restart_point (cnn,pkt);
	fin_or_rst_seen = 0;
	return (NULL);
    }
    return (pkt);
}


struct tcptb_pkt *tcptb_first_pkt (struct tcptb_cnn *cnn,
				   struct tcptb_cnn_stats *stats, 
				   struct tcptb_pkt_info *pkti)
{
    if ((cnn == NULL) || (stats == NULL))
    {
	tcptb_set_error ("no cnn or stats record given",FALSE);
	return (NULL);
    }
    if ((cnn == tcptb_restart_cnn) && (tcptb_restart_pkt != NULL))
	tcptb_current_pkt = tcptb_restart_pkt;
    else if (cnn->pkts_head != NULL)
    {
	tcptb_current_pkt = cnn->pkts_head;
	tcptb_restart_cnn = NULL;
	tcptb_restart_pkt = NULL;
    }
    else
    {
	tcptb_set_error ("no packets in connection",FALSE);
	tcptb_current_pkt = NULL;
	return (NULL);
    }
    bzero (stats,sizeof (struct tcptb_cnn_stats));
    stats->start_time = tcptb_current_pkt->ts.tv_sec + 
	(tcptb_current_pkt->ts.tv_usec / 1000000.);
    tcptb_stats_accum (cnn,tcptb_current_pkt,stats,pkti);
    return (tcptb_current_pkt);
}


struct tcptb_pkt *tcptb_next_pkt (struct tcptb_cnn *cnn, 
				  struct tcptb_cnn_stats *stats,
				  struct tcptb_pkt_info *pkti)
{
    if ((cnn == NULL) || (stats == NULL))
    {
	tcptb_set_error ("no connection or stats record given",FALSE);
	return (NULL);
    }
    if (tcptb_current_pkt != NULL)
	tcptb_current_pkt = (*tcptb_next_pkt_func) (cnn,tcptb_current_pkt);
    if (tcptb_current_pkt == NULL)
	return (NULL);
    tcptb_stats_accum (cnn,tcptb_current_pkt,stats,pkti);
    return (tcptb_current_pkt);
}
