
/* tcptb
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2004 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#ifdef LINUX
#include <net/ethernet.h>
#endif /* LINUX */
#include <strings.h>
#include <pcap.h>
#include <time.h>
#include <sys/time.h>

#ifndef IPTOS_CE
#define IPTOS_CE 0x01
#endif /* IPTOS_CE */

#ifndef IPTOS_ECT
#define IPTOS_ECT 0x02
#endif /* IPTOS_ECT */

#ifndef TH_ECE
#define TH_ECE 0x40
#endif /* TH_ECE */

#ifndef TH_CWR
#define TH_CWR 0x80
#endif /* TH_CWR */

#define TCPTB_ERR_SIZE        PCAP_ERRBUF_SIZE

/* These constants come from the size of the option space in the 
   TCP header. */
#define TCPTB_MAX_OPTS        40
#define TCPTB_MAX_SACK_BLOCKS 4

/* Handy TCP filters */
#define TCPTB_FILTER_TFR      "tcp[13] & 0x7 != 0"

struct tcptb_pkt
{
    struct timeval ts;
    unsigned short caplen;
    unsigned short len;
    char *pkt;
    struct tcptb_pkt *next;
};

struct tcptb_cnn
{
    unsigned int ip [2];
    unsigned short port [2];
    struct tcptb_pkt *pkts_head, *pkts_tail;
    struct tcptb_cnn *next;
};

struct tcptb_tcp_sack_block
{
    unsigned int start;
    unsigned int end;
};

struct tcptb_tcp_sack_opt
{
    unsigned short num_blocks;
    unsigned short dsack;
    struct tcptb_tcp_sack_block blk [TCPTB_MAX_SACK_BLOCKS];
};

struct tcptb_tcp_opts
{
    unsigned short num_opts;
    unsigned short ilformed;
    unsigned short partial;
    char *opts [TCPTB_MAX_OPTS];
};

struct tcptb_pkt_info
{
    double now;
    unsigned short payload_len;
    unsigned short endpoint;
    unsigned short other_endpoint;
    struct ip *iph;
    struct tcphdr *tcph;
    struct tcptb_tcp_opts opts;
    unsigned int end_seqno;
    struct tcptb_tcp_sack_opt sack;
};

struct tcptb_cnn_stats
{
    double start_time, last_time;
    unsigned long tot_pkts_sent [2];
    unsigned short syns_seen [2];
    unsigned char sack_ok [2];
    unsigned char ws_seen [2];
    unsigned char ws [2];
    unsigned short mss [2];
};

/* api.c */

char *tcptb_version ();
char *tcptb_error ();
int tcptb_open (char *filename);
void tcptb_close ();
int tcptb_slurp_trace ();
struct tcptb_cnn *tcptb_next_cnn ();
int tcptb_free_cnn (struct tcptb_cnn *cnn);
void tcptb_save_whole_packet (unsigned short flag);
void tcptb_save_ether_hdr (unsigned short flag);
int tcptb_set_hash_table_size (unsigned long sz);
int tcptb_set_bpf_filter (char *filter);

/* cnn_db.c */

#ifdef DEBUG
void tcptb_dump_hash_table_stats ();
#endif /* DEBUG */

/* cnn_iterator.c */

void tcptb_set_restart_point (struct tcptb_cnn *cnn, struct tcptb_pkt *pkt);
struct tcptb_cnn *tcptb_get_restart_cnn ();
struct tcptb_pkt *tcptb_get_restart_pkt ();
void tcptb_set_next_pkt_function (struct tcptb_pkt *(*func)());

struct tcptb_pkt *tcptb_first_pkt (struct tcptb_cnn *cnn,
				   struct tcptb_cnn_stats *stats, 
				   struct tcptb_pkt_info *pkti);
struct tcptb_pkt *tcptb_next_pkt (struct tcptb_cnn *cnn, 
				  struct tcptb_cnn_stats *stats,
				  struct tcptb_pkt_info *pkti);

/* util.c */

unsigned char tcptb_is_syn ();
unsigned char tcptb_is_fin ();
unsigned char tcptb_is_rst ();
unsigned char tcptb_is_ack ();
unsigned char tcptb_is_psh ();
unsigned char tcptb_is_urg ();
unsigned char tcptb_is_ece ();
unsigned char tcptb_is_cwr ();
unsigned char tcptb_is_ect ();
unsigned char tcptb_is_ce ();
struct ip *tcptb_iphdr ();
struct tcphdr *tcptb_tcphdr ();
char *tcptb_payload ();
int tcptb_payload_len ();
int tcptb_find_tcp_opts ();
int tcptb_get_ts_values ();
int tcptb_get_mss_value ();
int tcptb_get_ws_value ();
int tcptb_get_sack_blocks ();
