
/* tcptb
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2004 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <tcptb.h>
#include <tcptb-int.h>

unsigned long tcptb_num_pkts = 0;
unsigned short tcptb_swapped = FALSE;
unsigned short tcptb_keep_whole_packet = FALSE;
unsigned short tcptb_keep_ether_hdr = FALSE;


void tcptb_convert_ether (ethh)
struct ether_header *ethh;
{
    ethh->ether_type = ntohs (ethh->ether_type);
}


void tcptb_convert_ip (iph)
struct ip *iph;
{
    iph->ip_len = ntohs (iph->ip_len);
    iph->ip_id = ntohs (iph->ip_id);
    iph->ip_off = ntohs (iph->ip_off);
    iph->ip_sum = ntohs (iph->ip_sum);
    /* The standard routines for playing with IP addresses assume they are
       in network byte order, so leave them alone. */
#ifdef SWAP_IP_ADDRS
    iph->ip_src.s_addr = ntohl (iph->ip_src.s_addr);
    iph->ip_dst.s_addr = ntohl (iph->ip_dst.s_addr);
#endif /* SWAP_IP_ADDRS */
}


void tcptb_convert_tcp (tcph)
struct tcphdr *tcph;
{
    tcph->th_sport = ntohs (tcph->th_sport);
    tcph->th_dport = ntohs (tcph->th_dport);
    tcph->th_seq = ntohl (tcph->th_seq);
    tcph->th_ack = ntohl (tcph->th_ack);
    tcph->th_win = ntohs (tcph->th_win);
    tcph->th_sum = ntohs (tcph->th_sum);
    tcph->th_urp = ntohs (tcph->th_urp);
}


int tcptb_read_pkts (dump)
pcap_t *dump;
{
    struct pcap_pkthdr hdr;
    struct ether_header *ethh;
    struct ip *iph;
    struct tcphdr *tcph;
    char *packet, *read_packet;
    unsigned short ip_len, tcp_len, needed_len;
    unsigned short offset = 0;

    if (dump == NULL)
    {
	tcptb_set_error ("no dump to read from",FALSE);
	return (-1);
    }
    while ((read_packet = (char *)pcap_next (dump,&hdr)) != NULL)
    {
	tcptb_num_pkts++;
	if (hdr.caplen < EH_SIZE)
	    continue;
	ethh = (struct ether_header *)read_packet;
	tcptb_convert_ether (ethh);
	/* ignore if not IP (shouldn't be possible via the packet filter,
	   but just to be sure...) */
	if (ethh->ether_type != ETHERTYPE_IP)  
	    continue;
	if (hdr.caplen < (EH_SIZE + sizeof (struct ip)))
	    continue;
	iph = (struct ip *)(read_packet + EH_SIZE);
	/* ignore if not TCP (shouldn't be possible via the packet filter,
	   but just to be sure...) */
	if (iph->ip_p != IPPROTO_TCP)
	    continue;
	ip_len = iph->ip_hl * 4;
	if (ip_len < 20)
	    continue;
	if (hdr.caplen < (EH_SIZE + ip_len + sizeof (struct tcphdr)))
	    continue;
	tcph = (struct tcphdr *)(read_packet + (EH_SIZE + ip_len));
	tcp_len = tcph->th_off * 4;
	if (tcp_len < 20)
	    continue;
	tcptb_convert_ip (iph);
	tcptb_convert_tcp (tcph);
	if (tcptb_keep_whole_packet && tcptb_keep_ether_hdr)
	    needed_len = hdr.caplen;
	else if (tcptb_keep_whole_packet && !tcptb_keep_ether_hdr)
	{
	    needed_len = hdr.caplen - EH_SIZE;
	    offset = EH_SIZE;
	}
	else if (!tcptb_keep_whole_packet && tcptb_keep_ether_hdr)
	    needed_len = EH_SIZE + ip_len + tcp_len;
	else
	{
	    needed_len = ip_len + tcp_len;
	    offset = EH_SIZE;
	}
	if ((packet = (char *)tcptb_malloc (needed_len)) == NULL)
	{
	    tcptb_set_error ("fatal error: no memory for packet",TRUE);
	    return (-1);
	}
	bcopy (read_packet + offset,packet,needed_len);
	hdr.caplen = needed_len;
	if (tcptb_keep_ether_hdr)
	{
	    iph = (struct ip *)(packet + EH_SIZE);
	    tcph = (struct tcphdr *)(packet + (EH_SIZE + ip_len));
	}
	else
	{
	    iph = (struct ip *)packet;
	    tcph = (struct tcphdr *)(packet + ip_len);
	}
	tcptb_process_pkt (&hdr,packet,iph,tcph);
	if (tcptb_fatal_error)
	    return (-1);
    }
    return (0);
}
