
/* tcptb
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2004 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcptb.h>
#include <tcptb-int.h>


unsigned char tcptb_is_syn (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_SYN) ? TRUE : FALSE);
}


unsigned char tcptb_is_fin (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_FIN) ? TRUE : FALSE);
}


unsigned char tcptb_is_rst (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_RST) ? TRUE : FALSE);
}


unsigned char tcptb_is_ack (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_ACK) ? TRUE : FALSE);
}


unsigned char tcptb_is_psh (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_PUSH) ? TRUE : FALSE);
}


unsigned char tcptb_is_urg (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_URG) ? TRUE : FALSE);
}


unsigned char tcptb_is_ece (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_ECE) ? TRUE : FALSE);
}


unsigned char tcptb_is_cwr (struct tcphdr *tcph)
{
    if (tcph == NULL)
	return (0);
    return ((tcph->th_flags & TH_CWR) ? TRUE : FALSE);
}


unsigned char tcptb_is_ect (struct ip *iph)
{
    if ((iph->ip_tos & IPTOS_CE) && !(iph->ip_tos & IPTOS_ECT))
	return (TRUE);
    if (!(iph->ip_tos & IPTOS_CE) && (iph->ip_tos & IPTOS_ECT))
	return (TRUE);
    return (FALSE);
}


unsigned char tcptb_is_ce (struct ip *iph)
{
    if ((iph->ip_tos & IPTOS_CE) && (iph->ip_tos & IPTOS_ECT))
	return (TRUE);
    return (FALSE);
}


struct ip *tcptb_iphdr (struct tcptb_pkt *p)
{
    if (p == NULL)
	return (NULL);
    if (!tcptb_keep_ether_hdr)
	return ((struct ip *)p->pkt);
    if (p->caplen > EH_SIZE)
	return ((struct ip *)(p->pkt + EH_SIZE));
    return (NULL);
}


struct tcphdr *tcptb_tcphdr (struct tcptb_pkt *p)
{
    struct ip *iph = tcptb_iphdr (p);
    unsigned short offset;

    if ((p == NULL) || (iph == NULL))
	return (NULL);
    offset = ((tcptb_keep_ether_hdr) ? EH_SIZE : 0) + (iph->ip_hl * 4);
    if (p->caplen > offset)
	return ((struct tcphdr *)(p->pkt + offset));
    return (NULL);
}


char *tcptb_payload (struct tcptb_pkt *p)
{
    struct ip *iph = tcptb_iphdr (p);
    struct tcphdr *tcph = tcptb_tcphdr (p);
    unsigned short offset;

    if ((p == NULL) || (iph == NULL) || (tcph == NULL))
	return (NULL);
    offset = ((tcptb_keep_ether_hdr) ? EH_SIZE : 0) + (iph->ip_hl * 4) +
	(tcph->th_off * 4);
    if (p->caplen > offset)
	return ((char *)(p->pkt + offset));
    return (NULL);
}


int tcptb_payload_len (struct tcptb_pkt *p)
{
    struct ip *iph = tcptb_iphdr (p);
    struct tcphdr *tcph = tcptb_tcphdr (p);
    int len;

    if ((iph == NULL) || (tcph == NULL))
    {
	tcptb_set_error ("tcptb_payload_len: malformed packet",FALSE);
	return (-1);
    }
    len = iph->ip_len - (iph->ip_hl * 4) - (tcph->th_off * 4);
    return (len);
}


int tcptb_find_tcp_opts (struct tcphdr *tcph, struct tcptb_tcp_opts *opts)
{
    char *opt, *optend;
    unsigned short optlen;
    unsigned short next = 0;

    if ((tcph == NULL) || (opts == NULL))
    {
	tcptb_set_error ("tcptb_find_tcp_opts: NULL argument(s)",FALSE);
	return (-1);
    }
    bzero (opts,sizeof (struct tcptb_tcp_opts));
    opt = (char *)(tcph + 1);
    optend = (char *)tcph + (tcph->th_off * 4);
    while (opt < optend) 
    {
	opts->opts [next++] = opt;
	if ((opt + 1) == optend)
	{
	    opts->partial = 1;
	    break;
	}
	optlen = *(opt + 1);
	if (*opt == 0)
	    break;
	else if (*opt == 1)
	    opt++;
	else if (optlen < 2)
	{
	    opts->ilformed = 1;
	    break;
	}
	else if ((opt + optlen) <= optend)
	    opt += *(opt + 1);
	else
	{
	    opts->partial = 1;
	    break;
	}
    }
    opts->num_opts = next;
    /* If the last option we found is partial then reduce the number of
       options reported.  I.e., num_opts indicates the number of full 
       options present in the packet.  The partial option can still be 
       accessed and the "partial" field can be used to determine whether
       there is a partial option present or not. */
    if (opts->partial)
	opts->num_opts--;
    return (0);
}


int tcptb_get_ts_values (char *opt, unsigned long *ts_val, 
			 unsigned long *ts_ecr)
{
    if (*opt != TCPOPT_TIMESTAMP)
    {
	tcptb_set_error ("tcptb_get_ts_values: not a TS option",FALSE);
	return (-1);
    }
    if (*(opt + 1) != TCPOLEN_TIMESTAMP)
    {
	/* the TS option should have a length of 10 per RFC 1323 */
	tcptb_set_error ("tcptb_get_ts_values: wrong TS option length",FALSE);
	return (-2);
    }
    if ((ts_val == NULL) && (ts_ecr == NULL))
    {
	tcptb_set_error ("tcptb_get_ts_values: TS destinations both NULL",
			 FALSE);
	return (-3);
    }
    if (ts_val != NULL)
    {
	memcpy (ts_val,opt + 2,4);
	*ts_val = ntohl (*ts_val);
    }
    if (ts_ecr != NULL)
    {
	memcpy (ts_ecr,opt + 6,4);
	*ts_ecr = ntohl (*ts_ecr);
    }
    return (0);
}


int tcptb_get_mss_value (char *opt, unsigned short *mss)
{
    if (*opt != TCPOPT_MAXSEG)
    {
	tcptb_set_error ("tcptb_get_mss_value: not an MSS option",FALSE);
	return (-1);
    }
    if (*(opt + 1) != TCPOLEN_MAXSEG)
    {
	tcptb_set_error ("tcptb_get_mss_value: wrong MSS option length",FALSE);
	return (-2);
    }
    if (mss == NULL)
    {
	tcptb_set_error ("tcptb_get_mss_value: MSS desination is NULL",FALSE);
	return (-3);
    }
    memcpy (mss,opt + 2,2);
    *mss = ntohs (*mss);
    return (0);
}


int tcptb_get_ws_value (char *opt, unsigned char *ws)
{
    if (*opt != TCPOPT_WINDOW)
    {
	tcptb_set_error ("tcptb_get_ws_value: not a WS option",FALSE);
	return (-1);
    }
    if (*(opt + 1) != TCPOLEN_WINDOW)
    {
	tcptb_set_error ("tcptb_get_ws_value: wrong WS option length",FALSE);
	return (-2);
    }
    if (ws == NULL)
    {
	tcptb_set_error ("tcptb_get_ws_value: WS desination is NULL",FALSE);
	return (-3);
    }
    memcpy (ws,opt + 2,1);
    return (0);
}


int tcptb_get_sack_blocks (char *opt, struct tcptb_tcp_sack_opt *sack, 
			   unsigned int cumack)
{
    unsigned short i, offset;

    if (*opt != TCPOPT_SACK)
    {
	tcptb_set_error ("tcptb_get_sack_blocks: not a SACK option",FALSE);
	return (-1);
    }
    /* SACK option (from RFC 2018): 
         1 byte for kind
         1 byte for length
	 8 bytes per SACK block
     */
    if (*(opt + 1) < 10)
    {
	tcptb_set_error ("tcptb_get_sack_blocks: SACK option too short",FALSE);
	return (-2);
    }
    if (sack == NULL)
    {
	tcptb_set_error ("tcptb_get_sack_blocks: SACK block destination NULL",
			 FALSE);
	return (-3);
    }
    sack->num_blocks = (*(opt + 1) - 2) / 8;
    for (i = 0, offset = 2; i < sack->num_blocks; i++, offset += 8)
    {
	memcpy (&sack->blk [i].start,opt + offset,4);
	memcpy (&sack->blk [i].end,opt + offset + 4,4);
	sack->blk [i].start = ntohl (sack->blk [i].start);
	sack->blk [i].end = ntohl (sack->blk [i].end);
    }
    if (sack->blk [0].end <= cumack)
    {
	sack->dsack = TRUE;
	return (0);
    }
    for (i = 1; i < sack->num_blocks; i++)
	if ((sack->blk [0].start <= sack->blk [i].start) &&
	    (sack->blk [0].end >= sack->blk [i].end))
	{
	    sack->dsack = TRUE;
	    return (0);
	}
    return (0);
}

/* TODO: still need routines that will help process WS and maybe T/TCP */
