
/* tcpsum
 * Mark Allman (mallman@icir.org)
 *
 * Copyright (c) 2006 International Computer Science Institute
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcptb.h>

#define HISTORY_LEN  1024
#define SYN_PKT      1
#define FIN_PKT      2
#define RST_PKT      4
#define DAT_PKT      8
#define ACK_PKT      16

#define FALSE        0
#define TRUE         1

char history [HISTORY_LEN];


void history_append (code)
char code;
{
    unsigned short len;

    len = strlen (history);
    if (len > (HISTORY_LEN - 3))
	return;
    history [len++] = code;
    if (len == (HISTORY_LEN - 2))
	history [len++] = '+';
    history [len] = 0x0;
}


void process_one_connection (cnn)
struct tcptb_cnn *cnn;
{
    struct tcptb_pkt *pkt;
    struct tcptb_cnn_stats stats;
    struct tcptb_pkt_info pkti;
    unsigned long bytes [2];
    unsigned short bits_set;
    unsigned int syn_seq [2];
    unsigned int fin_seq [2];
    unsigned int rst_seq [2];
    unsigned int low [2];
    unsigned int high [2];
    unsigned int seen [2];
    unsigned int this_fin_seq;
    unsigned short i;
    unsigned short ep;
    unsigned short seqno_seen [2];
    unsigned char *p;

    history [0] = 0x0;
    seen [0] = seen [1] = 0;
    bytes [0] = bytes [1] = 0;
    seqno_seen [0] = seqno_seen [1] = FALSE;
    for (pkt = tcptb_first_pkt (cnn,&stats,&pkti); pkt != NULL; 
	 pkt = tcptb_next_pkt (cnn,&stats,&pkti))
    {
	ep = pkti.endpoint;
	bytes [ep] += pkti.payload_len;
        if (!seqno_seen [ep])
        {
            seqno_seen [ep] = TRUE;
            low [ep] = pkti.tcph->th_seq;
            high [ep] = pkti.tcph->th_seq + pkti.payload_len - 1;
        }
        else
        {
            if (pkti.tcph->th_seq < low [ep])
                low [ep] = pkti.tcph->th_seq;
            if ((pkti.tcph->th_seq + pkti.payload_len - 1) > high [ep])
                high [ep] = pkti.tcph->th_seq + pkti.payload_len - 1;
        }
	bits_set = tcptb_is_syn (pkti.tcph) + tcptb_is_fin (pkti.tcph) + 
	    tcptb_is_rst (pkti.tcph);
	if (bits_set > 1)
	    history_append (pkti.endpoint ? 'i' : 'I');
	else if (bits_set > 0)
	{
	    this_fin_seq = pkti.tcph->th_seq + pkti.payload_len;
	    if (tcptb_is_syn (pkti.tcph) && 
		(!(seen [ep] & SYN_PKT) || 
		 (pkti.tcph->th_seq != syn_seq [ep])))
	    {
		if (tcptb_is_ack (pkti.tcph))
		    history_append (pkti.endpoint ? 'h' : 'H');
		else
		    history_append (pkti.endpoint ? 's' : 'S');
		seen [ep] |= SYN_PKT;
		syn_seq [ep] = pkti.tcph->th_seq;
	    }
	    else if (tcptb_is_fin (pkti.tcph) && 
		     (!(seen [ep] & FIN_PKT) || 
		      (this_fin_seq != fin_seq [ep])))
	    {
		history_append (pkti.endpoint ? 'f' : 'F');
		seen [ep] |= FIN_PKT;
		fin_seq [ep] = this_fin_seq;
	    }
	    else if (tcptb_is_rst (pkti.tcph) && 
		     (!(seen [ep] & RST_PKT) || 
		      (pkti.tcph->th_seq != rst_seq [ep])))
	    {
		history_append (pkti.endpoint ? 'r' : 'R');
		seen [ep] |= RST_PKT;
		rst_seq [ep] = pkti.tcph->th_seq;
	    }
	}
	else 
	{
	    if (!(seen [ep] & ACK_PKT) && tcptb_is_ack (pkti.tcph) && 
		(pkti.payload_len == 0))
	    {
		history_append (pkti.endpoint ? 'a' : 'A');
		seen [ep] |= ACK_PKT;
	    }
	    else if (!(seen [ep] & DAT_PKT) && (pkti.payload_len > 0))
	    {
		history_append (pkti.endpoint ? 'd' : 'D');
		seen [ep] |= DAT_PKT;
	    }
	}
    }
    fprintf (stdout,"%0.6f %0.6f ", stats.start_time, 
	     stats.last_time - stats.start_time);
    for (i = 0; i < 2; i++)
    {
	p = (unsigned char *)&cnn->ip [i];
	fprintf (stdout,"%d.%d.%d.%d/%d ", p[0], p[1], p[2], p[3], 
		 cnn->port [i]);
	fprintf (stdout,"%ld %ld ", stats.tot_pkts_sent [i], bytes [i]);
        if (seqno_seen [i])
            fprintf (stdout,"%d ", high [i] - low [i]);
        else
            fprintf (stdout,"? ");
    }
    fprintf (stdout,"%s\n", strlen (history) ? history : "-");
}


void process_all_connections (trace_name)
char *trace_name;
{
    struct tcptb_cnn *cnn;

    for (cnn = tcptb_next_cnn (); cnn != NULL; cnn = tcptb_next_cnn ())
    {
        process_one_connection (cnn);
	tcptb_free_cnn (cnn);
    }
}


int main (argc,argv)
int argc; 
char *argv [];
{
    unsigned short i;

    if (argc < 2)
    {
	fprintf (stderr,"usage: tcpsum trace_file(s)\n");
	exit (1);
    }
    for (i = 1; i < argc; i++)
    {
	if (tcptb_open (argv [i]) < 0)
	{
	    fprintf (stdout,"%s\n", tcptb_error ());
	    exit (1);
	}
	tcptb_slurp_trace ();
	tcptb_close ();
    }
    process_all_connections (argv [1]);
    exit (0);
}
