/*
 * tcptsg - Copyright (c) 2004 Ethan Blanton, All Rights Reserved
 *
 * A simple application using mallman's TCP Toolbox to generate TCP
 * time-sequence graphs in the style of tcptrace.
 */
#define VERSION "1.0.2"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <tcptb.h>

#define MAX_NAMELEN 30
#define SEQ_CMP(x,y) ((signed)(x) - (signed)(y))

void plot_cnn_xpl (struct tcptb_cnn *cnn);

int main (int argc, char *argv[])
{
    struct tcptb_cnn *cnn;
    int c, i;

    while ((c = getopt (argc, argv, "vf:")) != -1) {
        switch (c) {
        case 'v':
            fprintf (stderr, "tcptsg version " VERSION "\n");
            return 0;
        case 'f':
            if (tcptb_set_bpf_filter (optarg)) {
                fprintf (stderr, "tcptsg: %s\n", tcptb_error ());
                return 1;
            }
            break;
        default:
            fprintf (stderr, "tcptsg: unrecognized option %c\n", c);
            return 1;
        }
    }
        
    if (argc - optind == 0) {
        fprintf (stderr, "usage: tcptsg -v\n");
        fprintf (stderr, "       tcptsg [-f filter] <tracefile> <...>\n");
        return 1;
    }

    for (i = optind; i < argc; i++) {
        if (tcptb_open (argv[i])) {
            fprintf (stderr, "tcptsg: %s\n", tcptb_error ());
            return 1;
        }

        while ((cnn = tcptb_next_cnn ()) != NULL) {
            plot_cnn_xpl (cnn);
            tcptb_free_cnn (cnn);
        }
        tcptb_close ();
    }

    return 0;
}

void plot_cnn_xpl (struct tcptb_cnn *cnn)
{
    char filename[MAX_NAMELEN];
    static int filen = 0;
    struct tcptb_cnn_stats stats;
    struct tcptb_pkt_info pkti;
    struct tcptb_pkt *pkt;
    char srcaddr[INET_ADDRSTRLEN], dstaddr[INET_ADDRSTRLEN];
    struct {
        FILE *fp;
        int ack_seen, ack;
        double ack_time;
        int win_seen, win;
        double win_time;
        int data_seen, data;
    } dir[2];

    memset (&dir[0], 0, sizeof(dir[0]));
    memset (&dir[1], 0, sizeof(dir[1]));

    sprintf (filename, "connection_%d_a.xpl", filen);
    if ((dir[0].fp = fopen (filename, "w+")) == NULL) {
        fprintf (stderr, "tcptsg: error opening output: %s\n",
                 strerror (errno));
        return;
    }
    sprintf (filename, "connection_%d_b.xpl", filen);
    if ((dir[1].fp = fopen (filename, "w+")) == NULL) {
        fprintf (stderr, "tcptsg: error opening output: %s\n",
                 strerror (errno));
        fclose (dir[0].fp);
        return;
    }
    filen++;

    inet_ntop (AF_INET, &cnn->ip[0], srcaddr, sizeof(srcaddr));
    inet_ntop (AF_INET, &cnn->ip[1], dstaddr, sizeof(dstaddr));
    
    fprintf (dir[0].fp, "timeval double\n");
    fprintf (dir[0].fp, "title\n%s:%hu -> %s:%hu (time-sequence graph)\n",
             srcaddr, cnn->port[0], dstaddr, cnn->port[1]);
    fprintf (dir[0].fp, "xlabel\ntime\n");
    fprintf (dir[0].fp, "ylabel\nsequence number\n");

    fprintf (dir[1].fp, "timeval double\n");
    fprintf (dir[1].fp, "title\n%s:%hu -> %s:%hu (time-sequence graph)\n",
             dstaddr, cnn->port[1], srcaddr, cnn->port[0]);
    fprintf (dir[1].fp, "xlabel\ntime\n");
    fprintf (dir[1].fp, "ylabel\nsequence number\n");

    for (pkt = tcptb_first_pkt (cnn, &stats, &pkti); pkt != NULL;
         pkt = tcptb_next_pkt (cnn, &stats, &pkti)) {
        struct ip *iph;
        struct tcphdr *tcph;
        unsigned int seq, dataseq;
        int len, s, d, i;

        iph = tcptb_iphdr (pkt);
        tcph = tcptb_tcphdr (pkt);
        seq = tcph->th_seq;
        dataseq = tcph->th_seq + (tcptb_is_syn (tcph) ? 1 : 0);
        len = tcptb_payload_len (pkt);

        s = (iph->ip_src.s_addr == cnn->ip[0]
             && tcph->th_sport == cnn->port[0]) ? 0 : 1;
        d = (s == 0) ? 1 : 0;

        if (tcptb_is_rst (tcph)) {
            fprintf (dir[s].fp, "magenta\n");
            fprintf (dir[s].fp, "x %f %u\n", pkti.now, seq);
            fprintf (dir[s].fp, "atext %f %u\nRST_OUT\n", pkti.now, seq);
            fprintf (dir[d].fp, "magenta\n");
            fprintf (dir[d].fp, "x %f %u\n", pkti.now, tcph->th_ack);
            fprintf (dir[d].fp, "atext %f %u\nRST_IN\n", pkti.now, tcph->th_ack);
            continue;
        }

        /* This packet */
        if (tcptb_is_syn (tcph)) {
            fprintf (dir[s].fp, "orange\n");
            fprintf (dir[s].fp, "diamond %f %u\n", pkti.now, seq);
            fprintf (dir[s].fp, "line %f %u %f %u\n", pkti.now, seq,
                     pkti.now, seq + 1);
            if (!len)
                fprintf (dir[s].fp, "uarrow %f %u\n", pkti.now, seq + 1);
            fprintf (dir[s].fp, "atext %f %u\nSYN\n", pkti.now, seq + len + 1);
        }
        if (len || tcptb_is_ack (tcph)) {
            if (dir[s].data_seen && SEQ_CMP(seq, dir[s].data) < 0) {
                fprintf (dir[s].fp, "red\n");
                fprintf (dir[s].fp, "atext %f %u\nR\n", pkti.now, dataseq + len);
            } else {
                fprintf (dir[s].fp, "white\n");
            }
            if (!tcptb_is_syn (tcph) && (len || !tcptb_is_fin (tcph)))
                fprintf (dir[s].fp, "darrow %f %u\n", pkti.now, dataseq);
            if (!tcptb_is_fin (tcph)) {
                if (tcptb_is_psh (tcph))
                    fprintf (dir[s].fp, "diamond %f %u\n", pkti.now,
                             dataseq + len);
                else
                    fprintf (dir[s].fp, "uarrow %f %u\n", pkti.now,
                             dataseq + len);
            }
            if (len)
                fprintf (dir[s].fp, "line %f %u %f %u\n", pkti.now, dataseq,
                         pkti.now, dataseq + len);
        }
        if (tcptb_is_fin (tcph)) {
            fprintf (dir[s].fp, "orange\n");
            if (!len)
                fprintf (dir[s].fp, "darrow %f %u\n", pkti.now, dataseq + len);
            fprintf (dir[s].fp, "line %f %u %f %u\n", pkti.now, dataseq + len,
                     pkti.now, dataseq + len + 1);
            fprintf (dir[s].fp, "box %f %u\n", pkti.now, dataseq + len + 1);
            fprintf (dir[s].fp, "atext %f %u\nFIN\n", pkti.now,
                     dataseq + len + 1);
        }
        if (dir[s].data_seen) {
            if (SEQ_CMP(pkti.end_seqno, dir[s].data) > 0)
                dir[s].data = pkti.end_seqno;
        } else {
            dir[s].data_seen = 1;
            dir[s].data = pkti.end_seqno;
        }

        /* The ACK line */
        if (dir[s].ack_seen && tcptb_is_ack (tcph)) {
            fprintf (dir[d].fp, "green\n");
            fprintf (dir[d].fp, "line %f %u %f %u\n", dir[s].ack_time, dir[s].ack,
                     pkti.now, dir[s].ack);
            if (SEQ_CMP(dir[s].ack, tcph->th_ack) == 0)
                fprintf (dir[d].fp, "dtick %f %u\n", pkti.now, tcph->th_ack);
            else
                fprintf (dir[d].fp, "line %f %u %f %u\n", pkti.now, dir[s].ack,
                         pkti.now, tcph->th_ack);
        }
        if (tcptb_is_ack (tcph)) {
            if (dir[s].ack_seen) {
                if (SEQ_CMP(dir[s].ack, tcph->th_ack) < 0) {
                    dir[s].ack = tcph->th_ack;
                    dir[s].ack_time = pkti.now;
                }
            } else {
                dir[s].ack_seen = 1;
                dir[s].ack = tcph->th_ack;
                dir[s].ack_time = pkti.now;
            }
        }

        /* SACK information */
        if (pkti.sack.num_blocks) {
            fprintf (dir[d].fp, "purple\n");
            for (i = 0; i < pkti.sack.num_blocks; i++) {
                fprintf (dir[d].fp, "htick %f %u\n", pkti.now,
                         pkti.sack.blk[i].start);
                fprintf (dir[d].fp, "line %f %u %f %u\n", pkti.now,
                         pkti.sack.blk[i].start, pkti.now,
                         pkti.sack.blk[i].end);
                fprintf (dir[d].fp, "htick %f %u\n", pkti.now,
                         pkti.sack.blk[i].end);
                fprintf (dir[d].fp, "atext %f %u\nS\n", pkti.now,
                         pkti.sack.blk[i].end);
                if (pkti.sack.num_blocks > 1)
                    fprintf (dir[d].fp, "rtext %f %u\n%d\n", pkti.now,
                             pkti.sack.blk[i].end, i + 1);
            }
        }

        /* Advertised window */
        if (tcptb_is_ack (tcph) && dir[s].win_seen) {
            fprintf (dir[d].fp, "yellow\n");
            fprintf (dir[d].fp, "line %f %u %f %u\n", dir[s].win_time,
                     dir[s].win, pkti.now, dir[s].win);
            if (SEQ_CMP(dir[s].win, tcph->th_ack + tcph->th_win))
                fprintf (dir[d].fp, "line %f %u %f %u\n", pkti.now, dir[s].win,
                         pkti.now, tcph->th_ack + tcph->th_win);
            else
                fprintf (dir[d].fp, "utick %f %u\n", pkti.now,
                         tcph->th_ack + tcph->th_win);
        }
        if (tcptb_is_ack (tcph)) {
            dir[s].win_seen = 1;
            dir[s].win = tcph->th_ack + tcph->th_win;
            dir[s].win_time = pkti.now;
        }

        if (ferror (dir[0].fp) || ferror (dir[1].fp)) {
            fprintf (stderr, "tcptsg: error writing output\n");
            fclose (dir[0].fp);
            fclose (dir[1].fp);
            return;
        }
    }

    fprintf (dir[0].fp, "go\n");
    fprintf (dir[1].fp, "go\n");
    fclose (dir[0].fp);
    fclose (dir[1].fp);
}
